import unittest

class test_Manager(unittest.TestCase):
    def setUp(self):
        from apetype.tasks import TaskBase, SKIP, SKIPCACHE, InjectCopy, InjectItems
        from apetype.workers import Manager
        
        class Task(TaskBase):
            a: int = 5
            
            def ar(_, a) -> list:
                return [a*i for i in range(10)]
            
            def aplus10(_, ar: InjectItems) -> list:
                print('y', ar)
                return ar+10
            
            def amaal10(_, ar: InjectItems, aplus10: InjectItems) -> list:
                return ar*aplus10
            
        task = Task(parse=True)
        self.manager = Manager(task, max_workers_x_subtask=2)

    def tearDown(self):
        del self.manager

    def test_manager(self):
        self.manager.start()
        self.assertEqual(len(self.manager.task._output), 3)



import unittest

class test_ConfigBase(unittest.TestCase):
    def setUp(self):
        import apetype as at
        class DepDepSettings(at.ConfigBase):
            c: int
            c_b: float = 3
            
        class DepSettings(at.ConfigBase):
            depdep: DepDepSettings
            b: int
            b_b: float
            b_c: int = 4
            
        class Settings(at.ConfigBase):
            depconfig: DepSettings
            a: int
            a_b: float
            a_c: int = 4

        self.Settings = Settings

    def tearDown(self):
        del self.Settings

    def test_config_with_deps_wo_prefix(self):
        settings = self.Settings(parse=False, prefix=False)
        settings.parse_args([0,0,0,0,0])
        self.assertEqual(settings.depconfig.b, 0)
        self.assertEqual(settings.b, 0)

    def test_config_with_deps_with_prefix(self):
        settings = self.Settings(parse=False, prefix=True)
        settings.parse_args([0,0,0,0,0])
        self.assertEqual(settings.depconfig.b, 0)
        self.assertEqual(settings.depconfig_b, 0)



"""apetype pipelines module

--------------------------------------------------------------
WORK IN PROGRESS
This module is not yet part of the documentation, and as such
not yet fully implemented. Let me know if you would like to be
able to use it.
--------------------------------------------------------------

I have put a lot of philosophical thought into what is the difference
between a pipeline and a task. In the implementation of this package
they both implement the RunInterface: they are supposed to run something
and reach a state of completion at some point.

A Task (TaskBase), when it explicitly defines all its dependencies, can 
be considered a pipeline indirectly, and it is not necessary to wrap it 
in a Pipeline class (PbBase).

A Pipeline class is useful when some glue code is necesarry between
different tasks:
    - output and input of subsequent Tasks must be mapped
    - a Task must be applied to a range of samples

Todo:
    - extracting task settings

Example:
    >>> from apetype.pipelines import PbBase, step
    ... from apetype.tasks import TaskBase
    ... 
    ... class TaskA(TaskBase):
    ...     a: int = 1
    ... 
    ...     def out1(self) -> int:
    ...         return self.a**2
    ... 
    ... class TaskB(TaskBase):
    ...     dependence: TaskBase
    ...     b: int = 2
    ... 
    ...     def out2(self) -> int:
    ...         return self.dependence.a * self.b
    ... 
    ... class TaskC(TaskBase):
    ...     taskB: TaskB
    ... 
    ... class Pipeline(PbBase):
    ...     taskA: TaskA = STEP()
    ...     taskB: TaskB = STEP(input={'dependence': 'taskA'})
    ...     end:   TaskC = STEP() # will automatically serve taskB
    ... 
    ... pipeline = Pipeline()
    ... pipeline.run()

"""
from .tasks import RunInterface
from itertools import count

class StepBase(object):
    lastNumber = count(0)
    
    def __init__(self, number = None, input = {}, output = {}):
        self.number = number if number else next(self.lastNumber)
        self.input = input
        self.output = output

    def __call__(self, pipeline, taskName):
        # TODO check previous ran tasks, and see if they need to be added
        task = pipeline._tasks[taskName](parse=False)
        if not task.completed():
            task(**{k:pipeline._tasks[v] for k,v in self.input.items()})
            task.run()
            pipeline._tasks[taskName] = task

STEP = StepBase

class PbBase(RunInterface):
    def __init__(self):
        import typing
        import inspect
        self._tasks = typing.get_type_hints(self)
        self.steps = list(self._tasks)
        self._step_management = {
            k:v for k,v in vars(type(self)).items()
            if k in self._tasks
        }
        for step in self.steps:
            if step not in self._step_management:
                self._step_management[step] = STEP(number=-1)
        self._ran = False

    def run(self):
        self._run_started = True
        for step in self.steps:
            self._step_management[step](
                pipeline = self,
                taskName = step
            )
        self._ran = True

    def completed(self):
        return self._ran



"""apetype reports module

Defines a TaskReport class, that using the leopard Report class
automatically builds a report as the Task is running.

Task subtasks can inject 'print' which then ensures that any
print statements are included in the report

Example:
    >>> from apetype.reports import TaskReport, TaskSection
    ... class Report(TaskReport):
    ...     title = 'An experimental report'
    ...     outfile = '/tmp/exreport'
    ...
    ...     def section_1(_) -> TaskSection:
    ...         import pandas as pd
    ...         return [
    ...           ('tab1', pd.DataFrame({'a': [1, 2], 'b': ['c', 'd']})),
    ...           ('tab2', pd.DataFrame({'c': [1, 2], 'd': ['c', 'd']}))
    ...         ]
    ...
    ...     def section_2_with_figure(_) -> TaskSection:
    ...         import matplotlib.pyplot as plt
    ...         print('Where will this end up?')
    ...         fig, ax = plt.subplots()
    ...         ax.scatter(range(5),range(5))
    ...         return {
    ...           'figures':{'fig 1':fig},
    ...           'clearpage':True
    ...         }

"""

from .tasks import TaskBase, PrintInject, ReturnTypeInterface

class TaskReport(TaskBase, PrintInject):
    def run(self, *args, show=True, **kwargs):
        # leopard needs to have been installed
        # if not `pip install leopard`
        # args and kwargs passed to super run
        import leopard as lp
        import inspect
        self.report = lp.Report(
            title = self.title,
            author = self.author if hasattr(self, 'author') else None,
            outfile = self.outfile,
            intro = inspect.getdoc(self) or ''
        )
        # // for newlines
        self.report.settings['doubleslashnewline'] = True
        # call super run to generate output
        super().run(*args, **kwargs)
        # output the report
        self.report.outputPDF(show=show,geometry_options=True)

    def _cache_postprocess(self, previous_output):
        """_cache_postprocess defines logic to execute
        after loading cached results.

        For the ReportTask this is just adding
        cached sections to the current report.
        """
        import leopard as lp
        if isinstance(previous_output, lp.Section):
            self.report.sections.append(previous_output)
        return previous_output
        
class TaskSection(dict, ReturnTypeInterface):
    def __init__(self):
        ReturnTypeInterface.__init__(self, dict)

    def __call__(self, task, function, result):
        import inspect
        from collections import OrderedDict
        from matplotlib.figure import Figure
        import pandas as pd
        if isinstance(result, list):
            # Optionally result can be a list of tuples
            # figures and tables will then be extracted in order
            figures = OrderedDict([r for r in result if isinstance(r[1],Figure)])
            tables = OrderedDict(
                [r for r in result
                 if isinstance(r[1],pd.DataFrame)
                 or isinstance(r[1],pd.Series)]
            )
            result = dict([r for r in result if r[0] not in figures.keys() or tables.keys()])
            if figures: result['figures'] = figures
            if tables: result['tables'] = tables
        if 'title' not in result:
            result['title'] = function.replace('_', ' ').capitalize()
        if 'text' not in result and inspect.getdoc(task.__getattribute__(function)):
            result['text'] = inspect.getdoc(task.__getattribute__(function))
        if 'clearpage' not in result and hasattr(task, 'clearpage'):
            result['clearpage'] = task.clearpage
        if hasattr(task, '_printout'):
            if 'code' in result:
                result['code'] += task.print()
            else: result['code'] = task.print()
        section = self.postprocess(result)
        task.report.subs.append(section)
        return section
        
    def preprocess(self):
        pass

    def postprocess(self, result):
        import leopard as lp
        return lp.Section(**result)



import os
import sys
import multiprocessing as mp

if sys.platform.lower() == "win32":
    # Command to activate term colors in windows
    os.system('')

# ANSI codes for different styles
class termstyle():
    "Info: https://en.wikipedia.org/wiki/ANSI_escape_code"
    BLACK = '\033[30m'
    RED = '\033[31m'
    GREEN = '\033[32m'
    YELLOW = '\033[33m'
    BLUE = '\033[34m'
    MAGENTA = '\033[35m'
    CYAN = '\033[36m'
    WHITE = '\033[37m'
    UNDERLINE = '\033[4m'
    BOLD = '\033[1m'
    RESET = '\033[0m'

# Counting semaphore
class CountSemaphore(object):
    def __init__(self, value, manager):
        self._value = value
        self._semaphore = manager.BoundedSemaphore(value)
        self._condition = manager.Condition()
        #self._counter = mp.Value('i', 0)

    def acquire(self, procnr):
        #with self._counter.get_lock():
        #    self._setvalue = self._counter.value
        #    self._counter.value += 1
        # Above strategy failed due to racing
        if procnr >= self._value: return False
        else: return self._semaphore.acquire(False)

    def release(self):
        return self._semaphore.release()



"""Module for handling multiprocessing
Workers take a Task and check which subtasks
can be executed

If subtasks generate figures, matplotlib needs to
use a compatible backend. Tested backends: 'pdf'
The main process should therefore execute:

>>> import matplotlib
... matplotlib.use('pdf')

Todo:
    - connect to workers on other servers
    - distribute TaskBase depedencies over the workers

Example:
    >>> import apetype as at
    ... import apetype.workers
    ... import time, numpy as np
    ... 
    ... class Task(at.TaskBase):
    ...     a: int = 5
    ... 
    ...     def ar(_, a) -> np.ndarray:
    ...         return np.array([a*i for i in range(10)])
    ... 
    ...     def aplus10(_, ar: at.tasks.InjectItems) -> list:
    ...         print('y', ar)
    ...         time.sleep(2)
    ...         return ar+10
    ... 
    ...     def amaal10(_, ar, aplus10) -> np.ndarray:
    ...         return ar*aplus10
    ... 
    ... task = Task(parse=True)
    ... manager = at.workers.Manager(task, max_workers_x_subtask=2)
    ... manager.start()
"""
import os
import multiprocessing as mp
from .utils import CountSemaphore
from .configs import ConfigBase
from .tasks import RunInterface, InjectItems

class Manager(ConfigBase, RunInterface):
    workers: int = os.cpu_count()
    max_workers_x_subtask: int = workers
    
    def __init__(self, task, *args, **kwargs):
        # Init ConfigBase, as Manager does not have positionals parse=True
        super().__init__(parse=True)
        # Set Manager options passed as kwargs
        if args or kwargs: self(*args, **kwargs)
        self.task = task
        # task has to be instantiated, ideally also parsed
        assert hasattr(self.task, '_output') and hasattr(self.task, '_input')
        self.manager = mp.Manager()
        self.task._input = self.manager.dict()
        self.task._output = self.manager.dict()
        self.task._output_tmp = {
            k:self.manager.dict()
            for k in self.task._output_functions.keys()
            if self.hasinjectitems(k)
        }
        self.task._output_locks = {
            k: CountSemaphore(self.max_workers_x_subtask, self.manager)
            if self.hasinjectitems(k) else mp.Lock()
            for k in self.task._output_functions.keys()
        }
        self.task._output_events = {
            k: mp.Event()
            for k in self.task._output_functions.keys()
        }
        self.pool = [
            WorkerProcess(self.task, i)
            for i in range(self.workers)
        ]
        
    def start(self):
        for p in self.pool:
            p.start()
            
        # Wait for workers to finish
        for p in self.pool:
            p.join()

        # Turn input and output back to normal dicts
        self.task._input = dict(self.task._input)
        self.task._output = dict(self.task._output)
        # Close mp manager
        self.manager.shutdown()

    def hasinjectitems(self, subtask):
        for p in self.task._output_functions[subtask].parameters.values():
            if p.annotation is InjectItems: return True
        return False

    # RunInterface
    def run(self): # Usual task options not implemented
        return self.start()

    def completed(self):
        return self.task.completed()

    @property
    def _input(self):
        return self.task._input

    @property
    def _output(self):
        return self.task._output


class WorkerProcess(mp.Process):
    def __init__(self, task, number): #, queue=None):
        self.task = task
        self.number = number # used to divide subtask workload
        #self.queue = queue
        super().__init__()

    def run(self):
        for subtask in self.task._output_functions:
            lock = self.task._output_locks[subtask]
            if lock.acquire(False if isinstance(lock, mp.synchronize.Lock) else self.number):
                # Before running subtask, check if there
                # is output for all dependencies
                for dependency in self.task._output_functions[subtask].parameters.keys():
                    if dependency in self.task._output_events:
                        depevent = self.task._output_events[dependency]
                        depevent.wait()
                        # If InjectItems, register length
                        if self.task._output_functions[
                                subtask
                        ].parameters[dependency].annotation is InjectItems:
                            dynamic_injectitems_len = len(self.task._output[dependency])
                if isinstance(lock, mp.synchronize.Lock) and not subtask in self.task._output:
                    # Checking if not yet in output
                    # as due to lock racing conditions
                    # another process might have already
                    # generated the output
                    print(self.name, 'running', subtask)
                    self.task.run(subtask)
                    self.task._output_events[subtask].set()
                elif (
                        type(lock) is CountSemaphore and
                        not subtask in self.task._output and
                        self.number < dynamic_injectitems_len
                ):
                    print(self.name, 'running part of', subtask)
                    partial_result = self.task.run(
                            {subtask: range(self.number,dynamic_injectitems_len,lock._value)}, return_tmp=True
                    )
                    for k,v in partial_result.items():
                        # updating on the managed dict gave racing issues
                        self.task._output_tmp[subtask][k] = v
                    #TODO check result type in section beneath
                    # Select proc 0 to wait for subtask to finish (to avoid racing)
                    if self.number == 0 and lock._condition.acquire(False):
                        #import time
                        print(self.name, 'waiting for other processes in', subtask)
                        # Wait for all processes to finish their subtask part
                        while len(self.task._output_tmp[subtask]) < dynamic_injectitems_len:
                            lock._condition.wait(10)
                            #print(self.name, self.task._output_tmp[subtask])
                            #time.sleep(5)
                        self.task._output[subtask] = [
                            self.task._output_tmp[subtask][i]
                            for i in range(dynamic_injectitems_len)
                        ]
                        del self.task._output_tmp[subtask]
                        self.task._output_events[subtask].set()
                    else:
                        try: lock._condition.notify()
                        except RuntimeError: pass
                                
                lock.release()



"""apetype tests Module
Defines TestTask that allows running a full test on a TaskBase derived class.
To run the test, create a class that inherits first from TestTask and then
from the Task class to be tested. Initiate and run.

The tests Module further defines all the tests for testing apetype itself.

Example:
>>> from apetype.tests import TestTask, TaskBase
... import numpy as np
... class Task(TaskBase):
...     a: int = 5
...     b: int = 3
... 
...     def ar(_, a, b) -> np.ndarray:
...         return np.array([a,2,3])+b
...     def aplus10(_, ar) -> np.ndarray:
...         return ar+10
...     def amaal10(_, ar, aplus10) -> np.ndarray:
...         return ar*aplus10
... class TestCase(TestTask, Task):
...     a = 1
...     b = 1
...     ar = np.array([3,3,4])
...     aplus10 = np.array([3,3,4])
...     amaal10 = np.array([3,3,4])
... testcase = TestCase()
... testcase.run_test()
"""
import unittest
from ..tasks import TaskBase

class TestTask(unittest.TestCase):
    """Provide the TaskBase class to be tested either
    by creating a new class that first inherits from
    TestTask and then from the TaskBase, or initiate
    a TestTask instance passing the TaskBase class as
    an argument

    Args:
      TaskClass (None|TaskBase): class to be tested.
    """
    def __init__(self, *args, TaskClass=None, **kwargs):
        # Initialize unittest.TestCase
        super().__init__(*args, **kwargs)
            
        if TaskClass or (len(self.__class__.mro()) > 3 and
                issubclass(self.__class__.mro()[3], TaskBase)):
            self._task = (
                TaskClass if TaskClass else self.__class__.mro()[3]
            )(parse=False)
            
            # Gather defined subtask outputs
            self._expected_output = {
                subtask: self.__getattribute__(subtask)
                for subtask in self._task._output_functions
            }
            
            # Gather parameter settings
            self._task_param_input = {
                p[0].replace('-',''):self.__getattribute__(p[0].replace('-',''))
                for p in self._task._settings
                if hasattr(self, p[0].replace('-',''))
            }

    def setUp(self):
        if hasattr(self, '_task'):
            self._task(**self._task_param_input)
            
    def test_run(self):
        if not hasattr(self, '_task'): return
        
        # Run the task
        self._task.run()

        # Assert task output
        for subtask in self._expected_output:
            self.assertEqual(
                self._expected_output[subtask],
                self._task._output[subtask]
            )


# Example Task and TestCase (that does not fail)
# serving as an integration test for apetype
class ExampleTask(TaskBase):
    a: int = 5
    b: int = 3
    
    def aplusb(_, a, b) -> int:
        return a+b
    def amaalaplusb(_, a, aplusb) -> int:
        return a*aplusb

class ExampleTestCase(TestTask, ExampleTask):
    a = 1
    b = 1
    aplusb = 2
    amaalaplusb = 2




"""apetype configs module defines the core class ConfigBase
that is also used as the starting point for Tasks and 
ReportTasks, and wrapped in Pipelines.

Settings are defined as typed attributes on an inheriting ConfigBase class

Example:

    >>> import apetype as at
    ... class DepSettings(at.ConfigBase):
    ...     b: int
    ...     b_b: float
    ...     b_c: int = 4
    ...     
    ... class Settings(at.ConfigBase):
    ...     depconfig: DepSettings
    ...     a: int
    ...     a_b: float
    ...     a_c: int = 4
    ... 
    ... settings = Settings(parse=[1,1,1,1])
    ... print(settings.depconfig_b, settings.a, settings.depconfig.b)
    1 1 1

"""

import typing
from collections import OrderedDict

class ConfigBase(object):
    """Takes in a list of settings, which will be exposed
    as CLI arguments. Each settings tuple should have the
    following format:
    ('--name', keyword dict for the parser.add_argument function)

    The recommended way to build a ConfigBase object, is to
    inherit from it and define the `_setup` method
    (see SettingsBase._setup docstring)

    Args:
      parse (bool|list|dict): if True, already parse arguments.
        Can also be a list that will then be passed on as args,
        all list items are converted to strings.
      prefix (bool): Whether or not to add the attribute name as  prefix
        to dependency settings. Prefix setting is propagated to downstream
        dependencies.
    """

    def __init__(self, parse=True, prefix=True):
        self.groups = False
        self._prefix = prefix
        self._help = self.get_arg_docs()
        if self.dependencies:
            self._set_input()
        self._setup()
        if self.dependencies:
            self._setup_from_deps()
        self._make_call()
        self._make_parser()
        if parse:
            if isinstance(parse, bool):
                self.parse_args(None)
            elif isinstance(parse, list):
                self.parse_args([str(p) for p in parse])
            elif isinstance(parse, dict):
                self(**parse)
            else:
                raise ValueError('parse was not bool|list|dict')

    def __getitem__(self, key):
        return self.settings.__getattribute__(key)

    def __getattr__(self, name):
        if 'settings' in self.__dict__ and name in self.settings:
            return self.settings.__getattribute__(name)
        else:
            # Trick from pandas to ensure jedi completion works
            return object.__getattribute__(self, name)

    def __repr__(self):
        if 'settings' in self.__dict__:
            return str(self.settings)
        else:
            return f'<Uninitialised {self.__class__.__name__}>'

    def __str__(self):
        return self.__repr__().replace(
            'Namespace(', 'Settings:\n\n').replace(', ', '\n')[:-1]

    def _setup(self):
        """Can be overwritten by inheriting classes.
        Allows defining parameters with type hints.
        Overwritten _setup methods need to call `super()._setup()`
        at the end.

        Example:
        >>> class Settings(SettingsBase):
        ...     def _setup(_, a: int = 5, b: float = .1, c: str = 'a'):
        ...          super()._setup()
        ... settings = Settings()
        """
        import inspect
        cls = type(self)
        cls_vars = {}
        for c in cls.mro()[::-1]:
            if issubclass(c, ConfigBase) and c is not ConfigBase:
                cls_vars.update(vars(c))
        # getting hints from self does not always work
        self._annotations = annotations = typing.get_type_hints(cls)
        self._annotation_groups = annotation_groups = [
            attr for attr in cls_vars if isinstance(cls_vars[attr], type)
        ]
        if annotations or annotation_groups:
            # get_attr does not work so requesting vars
            self._settings = [
                self.add_arg_format(p, cls_vars[p] if p in cls_vars else None,
                                    annotations[p], p not in cls_vars,
                                    self._help[p] if p in self._help else None
                )
                for p in annotations
                if not issubclass(annotations[p], ConfigBase)
                # in future could make exception for parseable ConfigBase class
                # also still need solution for list|set|... annotations
            ]
            if annotation_groups:
                self.groups = True
                self._settings = OrderedDict(
                    ('default', self._settings)
                ) if self._settings else OrderedDict()
                for annot_grp in annotation_groups:
                    grp_annotations = typing.get_type_hints(cls_vars[annot_grp])
                    grp_cls_vars = vars(cls_vars[annot_grp])
                    self._settings[annot_grp] = [
                        self.add_arg_format(
                            p, grp_cls_vars[p] if p in grp_cls_vars else None,
                            grp_annotations[p], p not in grp_cls_vars,
                            self._help[p] if p in self._help else None
                        )
                        for p in grp_annotations
                    ]
        else:
            sig = inspect.signature(self._setup)
            self._settings = [
                self.add_arg_format(
                    p, sig.parameters[p].default,
                    sig.parameters[p].annotation,
                    not(bool(sig.parameters[p].default)),
                    self._help[p] if p in self._help else None
                )
                for p in sig.parameters
            ]

    def _setup_from_deps(self):
        for dependency in self._input:
            if self._prefix:
                depsettings = [
                    (
                        f'{dependency}_{s[0]}',
                        (
                            f'--{dependency}-{s[1][0][2:]}' if s[1][0].startswith('--')
                            else f'{dependency}_{s[1][0]}',
                            s[1][1]
                        )
                    )
                    for s in self._input[dependency]._settings
                ]
            else:
                depsettings = self._input[dependency]._settings
            self._settings += depsettings
        
    @staticmethod
    def add_arg_format(name, default, typ, positional, help):
        if typ is bool:
            return (name, (f'--{name.replace("_","-")}', {
                    'default': default,
                    'action': 'store_const',
                    'const': not(default),
                    'help': help
            }
            ))
        else:
            return (name, (name if positional else f'--{name.replace("_","-")}', {
                    'default': default,
                    'type': typ,
                    'help': help
            }
            ))

    @classmethod
    def get_arg_docs(cls):
        import re
        import inspect
        commentdoc = re.compile(r'\s*(\w+)\s*:\s*\w+.*?#(.+)')
        multilinedoc = re.compile(
            r'\s*(?P<identifier>\w+)\s*:\s*\w+.*?\n\s*(?P<delimiter>"""|\'\'\')'
            r'(?P<annot>(.*\n)+?)\s*?(?P=delimiter)',
            re.MULTILINE)
        onelinedoc = re.compile(
            r'\s*(?P<identifier>\w+)\s*:\s*\w+.*?\n\s*(?P<delimiter>"|\')'
            r'(?P<annot>.+)(?P=delimiter)')
        try:
            argdocs = dict([
                commentdoc.match(line).groups()
                for line in inspect.getsourcelines(cls)[0] if commentdoc.match(line)
            ])
            # Clean if not identifier and strip
            argdocs = {arg:argdocs[arg].strip() for arg in argdocs if arg.isidentifier()}
            argdocs.update({
                match.group('identifier'):match.group('annot').replace('\n', ' ').strip()
                for match in onelinedoc.finditer(inspect.getsource(cls))
                if match.group('identifier').isidentifier()
            })
            argdocs.update({
                match.group('identifier'):match.group('annot').replace('\n', ' ').strip()
                for match in multilinedoc.finditer(inspect.getsource(cls))
                if match.group('identifier').isidentifier()
            })
            return argdocs
        except TypeError:
            # TypeError will be raised when class is defined interactively
            # Returning an empty dict in that case
            return {}

    def _set_settings(self, vardict):
        from argparse import Namespace
        if 'self' in vardict: vardict.pop('self')
        for attr in vardict:
            # TODO check type
            self.__setattr__(attr, vardict[attr])
            # print(attr,vardict[attr])
        # To have consistent type of self.settings using Namespace
        # to match self.parse_args
        self.settings = Namespace(**vardict)
        if self.dependencies: self._set_input_settings(vardict)

    def _make_call(self):
        from types import FunctionType
        variables = ', '.join([
            s[0] + ('' if s[1][1]['default'] is None else
                                     '='+repr(s[1][1]['default'])
            )
            for s in sorted(
                self._settings,
                key=lambda x: x[1][1]['default'] is None,
                reverse=True
            )
        ])
        call_code = compile(f'''def set_variables(self, {variables}):
            self._set_settings(locals())
        ''', "<string>", "exec")
        # Find code sub object:
        #for i in range(len(call_code.co_consts)):
        #    if type(call_code) is type(call_code.co_consts[i]):
        #        code = call_code.co_consts[i]
        #        break
        code_pos = 0 if call_code.co_consts[-1] is None else -4
        call_func = FunctionType(
            call_code.co_consts[code_pos], globals(), "set_variables",
            argdefs = call_code.co_consts[-1] if '=' in variables else None
        )
        setattr(self.__class__, '__call__', call_func)
        
    def _make_parser(self, **kwargs):
        import argparse
        self.parser = argparse.ArgumentParser(**kwargs)
        if self.groups: self.group_parsers = {}
        for grp in self._settings:
            if self.groups:
                parser = self.parser.add_argument_group(grp)
                self.group_parsers[grp] = parser
            else:
                parser = self.parser
            for setting in (self._settings[grp] if self.groups else self._settings):
                parser.add_argument(
                    setting[1][0],
                    **setting[1][1]
                )
            if not self.groups:
                break  # if no groups need to break

    def parse_args(self, args):
        self.settings = self.parser.parse_args(args)
        # Overwrite default typing values
        for a in self._annotations:
            if a in self.settings:
                self.__setattr__(a, self.settings.__getattribute__(a))
        if self.dependencies: self._set_input_settings(vars(self.settings))
        return self.settings

    # Section for dealing with dependencies
    @classmethod
    def _get_dependencies(cls):
        """Recursive property that lists all dependencies.
        Returns None if no further dependencies.

        `_get_dependencies` works recursive, because the main ConfigBase object
        needs to know all the settings that have to be requested either through
        a command line or function call.
        """
        dependency_tree = []
        clsvars = cls.__annotations__
        for attr in clsvars:
            if (
                    isinstance(clsvars[attr], type) and
                    issubclass(clsvars[attr], ConfigBase)
            ):
                dependency = clsvars[attr]
                dependency_tree.append(
                    (attr, dependency, dependency._get_dependencies())
                )

        return dependency_tree if dependency_tree else None

    @property
    def dependencies(self):
        return self._get_dependencies()

    def _set_input(self):
        # Instatiate every dependency ConfigBase without parsing
        # this will be recursive as each ConfigBase will do the same for theirs
        self._input = {
            dep[0]:dep[1](parse=False, prefix=self._prefix)
            for dep in self.dependencies
        }
        # Make dependencies accesible as attribute to self ConfigBase instance
        for k,v in self._input.items():
            self.__setattr__(k, v)

    def _set_input_settings(self, vardict):
        for dependency in self._input:
            depsettingsset = {
                f'{dependency}_{s[0]}' if self._prefix else s[0]
                for s in self._input[dependency]._settings
            }
            self._input[dependency](
                **{
                    (k[len(dependency)+1:] if self._prefix else k):v
                    for k,v in vardict.items()
                    if k in depsettingsset
                }
            )



"""apetype tasks module takes the ConfigBase to build an inheritable TaskBase
class around it. Through using inheritance or including other TaskBase classes
as dependencies, the result is an implicit pipeline. For explicit pipelines
where the input and output of subsequent dependencies needs to be fitted, see
the `apetype.pipelines`.

In the example below, TaskBase dependencies is illustrated, and subtasks that
are executed in a subprocess with either a shell or python interpreter.
In the task subtasks, `self` is replaced with `_`. This is to make it clearer,
that these methods are usually not called directly by the end-user. However,
if so desired, `self` can be used. Furthermore, when writing tests, it can be
appropriate to call the subtask methods directly, allowing to control what is
being injected.

TODO:
  - for subtasks running in subprocesses, use the return_annotation to automate
    when python syntax can be used, or transformed to the syntax of the interpreter

Example:

    >>> import apetype as at
    ... class TaskDep(at.TaskBase):
    ...     a: str = '/tmp/file1'
    ... 
    ...     def generate_output(_, a) -> str:
    ...         return a    
    ... 
    ... class Task(at.TaskBase):    
    ...     # Task settings
    ...     a: int = 10
    ...     b: str = 'a'
    ... 
    ...     # Task dependencies
    ...     task_dependence1: TaskDep
    ... 
    ...     # Subtasks
    ...     def generate_output1(_, task_dependence1) -> int:
    ...         print(task_dependence1.a)
    ...         return 0
    ...     
    ...     def generate_output2(_) -> str:
    ...         with _.env('sh') as env:
    ...             env.exec('which python')
    ...             return env.output
    ... 
    ...     def generate_output3(_) -> str:
    ...         with _.env('py') as env:
    ...             env.exec(f'''
    ...             for i in range({_.a}):
    ...                 print(i)
    ...             ''')
    ...             return env.output
    ... 
    ... task = Task()
    ... task.run()
    ... print(task._input, task._output)

""" 

import os
import abc
import typing
import inspect
import pickle
from collections import OrderedDict
from .configs import ConfigBase

# Interface for tasks and pipelines
class RunInterface(abc.ABC):
    @abc.abstractmethod
    def run(self):
        pass

    @abc.abstractmethod
    def completed(self):
        pass

# Interfaces for logic to perform on subtask return values
class ReturnTypeInterface(abc.ABC):
    def __init__(self, type):
        self.type = type

    @abc.abstractmethod
    def __call__(self, result):
        pass
    
    @abc.abstractmethod
    def preprocess(self):
        pass

    @abc.abstractmethod
    def postprocess(self):
        pass

class SKIP(ReturnTypeInterface):
    # SKIP subtasks
    # if subtask is explicitly mentioned in run list
    # reset type to type SKIP was instantiated with
    def __call__(self, result):
        pass
    def preprocess(self):
        # Returns SKIP instance type
        return self.type
    def postprocess(self):
        pass

class SKIPCACHE(ReturnTypeInterface):
    # Could also consider a factory method for both
    # Inheritance does not work because isinstance needs to be separate
    def __call__(self, result):
        pass
    def preprocess(self):
        # Returns SKIP instance type
        return self.type
    def postprocess(self):
        pass

# Interfaces for logic to perform on subtask value injections
class InjectInterface(abc.ABC):
    """Any InjectInterface (II) has to define a __call__ routine
    that takes the parameter that was annotated with the II, the
    subtask name and a flag dict. Flags can be modified, but 
    __call__ should only return the transformed parameter.
    """
    @abc.abstractmethod
    def __call__(self, parameter, subtask, flags):
        pass

class InjectCopy(InjectInterface):
    """To avoid having side effects on injected parameters,
    objects such as a pd.DataFrame can use this to inject
    a copy instead
    """
    def __call__(self, parameter, subtask, flags):
        return parameter.copy()

class InjectItems(InjectInterface):
    """The passed parameter should be of type list, tuple, or generator.
    An enumerate of the parameter is returned.
    """
    def __call__(self, parameter, subtask, flags):
        try:
            flags['injectitems'].add(subtask)
            assert flags['injectitems_len'] == len(parameter)
        except KeyError:
            flags['injectitems'] = {subtask}
            flags['injectitems_len'] = len(parameter)
        except AssertionError:
            raise Exception('All InjectItems params dynamic lists need to have same size')
        return zip([subtask]*len(parameter), parameter)
    
# Base class for tasks
class TaskBase(ConfigBase, RunInterface):
    def __init__(self, parse=False, prefix=True, run=False):
        """
        Parsing at object creation is not that useful
        Only when a task runs should it have all its settings.
        When run is True, parse will also be set to True.

        Args:
          parse (bool): parse settings
          prefix (bool): prefix dependency tasks or configs with attribute name
          run (bool|list): run the task upon creation, can also
            be a list of subtasks to run.
        """
        super().__init__(parse=parse or bool(run), prefix=prefix)
        self._taskprep()
        if not self.dependencies:
            # subtasks might still contain dependencies that will be
            # stored in self._input
            self._input = {}
        self._output = {}
        if run:
            self.run(
                subtasks=None if isinstance(run,bool) else run
            )
        
    def _taskprep(self):
        cls = type(self)

        # Small utility function to sort members in order of appearance
        def memberline(m):
            try:
                return m[1].__code__.co_firstlineno
            except AttributeError:
                try:
                    # Interactive code may need extra attribute
                    return m[1].__func__.__code__.co_firstlineno
                except:
                    return -1
    
        self._output_functions = OrderedDict([
            (name, inspect.signature(fun))
            for name, fun in sorted(
                    inspect.getmembers(cls, predicate=inspect.isfunction),
                    key=memberline
            )
            if typing.get_type_hints(fun)
        ])

    def _subtask_dependencies(self, subtask):
        """Get the dependencies for the subtask.
        If one of the dependencies is not available yet,
        returns False. A dependency is always first looked
        up into the task's input dict, i.c. external dependencies,
        secondly in the output dict, lastly the task attributes.

        Args:
          subtask (str): method str of the task
        
        Returns:
          dict | False
        """
        parameters = self._output_functions[subtask].parameters
        function_inputs = {}
        flags = {} # can be set by InjectInterface
        for dependency in parameters:
            if dependency in self._input:
                function_inputs[dependency] = self._input[dependency]
            # If annotation present for dependency it should have RunInterface or InjectInterface
            elif parameters[dependency].annotation is not inspect._empty:
                annotation = parameters[dependency].annotation
                if issubclass(annotation, RunInterface):
                    # Instantiate dependency and run
                    deptask = annotation() # TODO handling settings subtask
                    deptask.run()
                    self._input[dependency] = deptask
                    function_inputs[dependency] = deptask
                elif issubclass(annotation, InjectInterface):
                    try: function_inputs[dependency] = annotation()(
                            self._output[dependency], dependency, flags
                    )
                    except KeyError:
                        if dependency in self._output_functions: return False
                        else:
                            print('Currently InjectInterface only on task subtask output')
                            raise
                else:
                    raise Exception(
                        'Only RunInterface or InjectInterface supported for param annotation'
                    )
            # If no annotation it could be the output generated from one of the task methods
            elif dependency in self._output:
                function_inputs[dependency] = self._output[dependency]
            # Finally, it could also be an attribute of the task class
            else:
                try:
                    function_inputs[dependency] = self.__getattribute__(dependency)
                except AttributeError:
                    # check if attribute simply refers to 'self' or similar
                    if dependency not in ('_', 'self', 'task'):
                        if dependency in self._output_functions:
                            return False
                        else:
                            print(dependency, 'not defined/found')
                            raise
        # Return function_inputs for executing
        if not flags:
            return function_inputs
        else:
            if 'injectitems' in flags:
                parameterlist = zip(*[function_inputs.pop(k) for k in flags['injectitems']])
                return (flags['injectitems_len'], (
                    (itemnr,{**function_inputs, **dict(injitems)})
                    for itemnr, injitems in enumerate(parameterlist)
                ))
            
    def run(self, subtasks=None, fail=True, load_cache=False, return_tmp=False):
        # Task can declare a verbose attribute used in this run
        try:
            from .utils import termstyle as ts
            verbose = self.__getattribute__('verbose')
        except AttributeError: verbose = False

        # First check if external dependencies completed
        if self.dependencies:
            self._run_task_dependencies()
        
        # If subtasks is a str (when only 1 subtask needs to be executed
        # redefine it as a list to make compatible with the code
        if isinstance(subtasks, str): subtasks = [subtasks]
        
        # Run task subtask methods
        for fn in self._output_functions:
            return_annotation = self._output_functions[fn].return_annotation
            
            # If subtasks are specified, only run those
            if subtasks and fn not in subtasks: continue
            elif isinstance(return_annotation, SKIP):
                if subtasks and fn in subtasks:
                    return_annotation = return_annotation.preprocess()
                else: continue
            # If subtask already generated output continue
            if fn in self._output:
                if verbose: print(f'{ts.GREEN}{fn}{ts.RESET}', 'already generated output')
                continue
            elif hasattr(self, 'cache'):
                # If cache is set as attribute check if previous output can be loaded
                cachefilename = os.path.join(
                    self.cache,
                    f'{fn}.pickle'
                )
                if load_cache and os.path.exists(cachefilename) and not isinstance(
                        return_annotation, SKIPCACHE):
                    self._output[fn] = pickle.load(open(cachefilename, 'rb'))
                    if hasattr(self, '_cache_postprocess'):
                        # Inheriting TaskBase classes can define logic for handling cache
                        self._output[fn] = self._cache_postprocess(self._output[fn])
                    if verbose: print(f'{ts.GREEN}{fn}{ts.RESET}', 'cached output loaded')
                    continue
            return_type = return_annotation if not isinstance(
                return_annotation, SKIPCACHE) else return_annotation.preprocess()

            # TODO could check here if return type is correct in earlier generated output
            if verbose: print(f'{ts.BOLD}Executing{ts.RESET}', f'{ts.GREEN}{fn}{ts.RESET}', '...')
            function_inputs = self._subtask_dependencies(fn)
            # Checking if function_inputs is False (empty is ok = no deps)
            if function_inputs is False:
                raise Exception(f'Unmet dependencies for {fn}')
            if isinstance(function_inputs, dict):
                # regular situation -> one function_inputs
                return_value = self.__getattribute__(fn)(
                    **function_inputs
                )
            else: # function_inputs is a generator
                from collections.abc import Iterable
                complete_output, function_inputs = function_inputs
                if isinstance(subtasks, dict) and isinstance(subtasks[fn], Iterable):
                    return_value = {
                        fi[0]: self.__getattribute__(fn)(**fi[1])
                        for fi in function_inputs
                        if fi[0] in subtasks[fn]
                    }
                    if return_tmp: return return_value
                    try:
                        self._output_tmp[fn].update(return_value)
                        # Check if tmp result is complete
                        if len(self._output_tmp[fn]) == complete_output:
                            return_value = [
                                self._output_tmp[fn][i]
                                for i in sorted(self._output_tmp[fn])
                            ]
                            del self._output_tmp[fn]
                    except AttributeError:
                        self._output_tmp = {fn: return_value}
                        return
                    except KeyError:
                        self._output_tmp[fn] = return_value
                        return
                else:
                    return_value = [
                        self.__getattribute__(fn)(**fi[1])
                        for fi in function_inputs
                    ]
            if verbose: print(f'{ts.BOLD}done{ts.RESET}')
            if issubclass(return_type, ReturnTypeInterface):
                return_instance = return_type()
                self._output[fn] = return_instance(
                    task = self,
                    function = fn,
                    result = return_value
                )
            else:
                assert isinstance(return_value, return_type)
                self._output[fn] = return_value
            if hasattr(self, 'cache') and not isinstance(return_annotation, SKIPCACHE):
                # Write out output in cache dir if provided
                pickle.dump(self._output[fn], open(cachefilename, 'wb'))

    def env(self, environment):
        return ExecEnvironment(environment)

    def completed(self):
        return bool(self._output)

    # Section for dealing with external task dependencies
    def _run_task_dependencies(self):
        # Will need to be overloaded for handling multiprocessing of deps
        for dependency, deptask in self._input.items():
            if isinstance(deptask, TaskBase) and not deptask.completed():
                deptask.run()
        
class PrintInject(object):
    """Mixin class to add print diverting logic to a task.
    Classes that inherit this class, next to TaskBase can
    inject print in the subtasks.
    """
    def print(self, *args, **kwargs):
        """Method that can be used instead of print, to
        capture the stdout.

        When called without args or kwargs, returns the
        current buffer and resets it.
        """
        from io import StringIO
        if not args and not kwargs:
            printout = self._printout
            del self._printout
            return printout
        elif not 'file' in kwargs:
            out = StringIO()
            print(*args, file=out, **kwargs)
            print(out.getvalue(), end='')
            try: self._printout += out.getvalue()
            except AttributeError: self._printout = out.getvalue()
        else: print(*args, **kwargs)
        
class ExecEnvironment(object):
    predefined_envs = {
        'sh': ['bash', []],
        'py': ['python', []]
    }
    
    def __init__(self, command, options=[]):
        import threading
        import shutil
        self.lock = threading.Lock()
        self.command = shutil.which(
            self.predefined_envs[command][0]
            if command in self.predefined_envs else command
        )
        self.options = (options if options else
            self.predefined_envs[command][1]
            if command in self.predefined_envs else []
        )

    def __enter__(self):
        import tempfile
        self.lock.acquire()
        self.tmpfile = tempfile.NamedTemporaryFile(mode = 'w+t', delete = False)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        # TODO log issues
        os.remove(self.tmpfile.name)
        self.lock.release()

    def exec(self, script, check = True, reident=True):
        import subprocess
        import re
        
        # reident
        if reident:
            identspace = re.compile(r'.*\n(\W*)\w')
            if identspace.match(script):
                # TODO should split over lines and then reassemble
                script = script.replace('\n'+identspace.match(script).groups()[0],'\n')
            
        # write tmp script file
        try:
            self.tmpfile.write(script)
        finally:
            self.tmpfile.close()
            
        # execute
        proc = subprocess.run(
            [self.command]+self.options+[self.tmpfile.name],
            text = True, capture_output = True, check = check
        )
        self.output = proc.stdout
        self.error = proc.stderr



import unittest

class test_TaskBase(unittest.TestCase):
    def setUp(self):
        import apetype as at
        class DepDepTask(at.TaskBase):
            c: int
            c_b: float = 3

            def sum(_, c, c_b) -> float:
                return .5+c+c_b
            
        class DepTask(at.TaskBase):
            depdep: DepDepTask
            b: int
            b_b: float
            b_c: int = 4

            def product(_, b, b_c) -> int:
                return b*b_c
            
        class Task(at.TaskBase):
            deptask: DepTask
            a: int
            a_b: float
            a_c: int = 4

            def product(_, a, deptask) -> int:
                return a*deptask._output['product']

        self.Task = Task

    def tearDown(self):
        del self.Task

    def test_task_with_deps_wo_prefix(self):
        task = self.Task(parse=False, prefix=False)
        task.parse_args([0,0,0,0,0])
        task.run()
        self.assertEqual(task.deptask.b, 0)
        self.assertEqual(task.b, 0)
        self.assertEqual(task._output['product'], 0)
        self.assertEqual(task.deptask._output['product'], 0)
        
    def test_task_with_deps_with_prefix(self):
        task = self.Task(parse=False, prefix=True)
        task.parse_args([0,0,0,0,0])
        task.run()
        self.assertEqual(task.deptask.b, 0)
        self.assertEqual(task.deptask_b, 0)
        self.assertEqual(task._output['product'], 0)
        self.assertEqual(task.deptask._output['product'], 0)



"""APEtype module documentation

`ConfigBase` is the starting point for anything APEtype. It is a base
class to use to turn any script into a library function or
command-line call. Simply derive a class from it, and add typed
attributes.  Those attributes will then be offered as parameters to
the call of the instances of this class, or if you pass `parse=True`
on instantiating, will be read from the command line.

    >>> import apetype as at
    ... import apetype.tests
    ... import apetype.workers
    ... 
    ... class MySettings(at.ConfigBase):
    ...     pos1: int # After comment shown as help
    ...     pos2: str
    ...     '''
    ...     .. or the docstr following the attribute
    ...     '''
    ...     kw1: int = 5
    ...     'with one or multiline docstr'
    ... 
    ... settings = MySettings(parse=False)
    ... print(settings.kw1)

To execute a program, 2 questions are important: "What should it do?"
and "What should it know to do that?". The latter is handled by
`ConfigBase`, the former is handled by the `TaskBase`, which contains
the same configuration options and derives from `ConfigBase`, but also
contains *to do* logic.  Any method that has type annotation is
considered a subtask, and when calling the task `run` method all
subtasks or a selection are executed in order. Any `ConfigBase`
attributes can be injected in a subtask method, or can be accessed as
attribute on the self object, which is denoted as `_` to differentiate
it from regular Python object methods.

    >>> class MyTask(at.TaskBase, MySettings):
    ...     kw2: int = 3
    ...     def addition(_, pos1, kw2) -> int:
    ...         return pos1+kw2
    ...     def multiply(_, pos2, kw1) -> str:
    ...         return pos2*(kw1+_.kw2)
    ... mytask = MyTask(parse={'pos1':1,'pos2':'1'})
    ... mytask.run()
    ... print(mytask._output['addition'], mytask._output['multiply'])

Use all your system's processing capacity by delegating the task
to a manager:

    >>> mpTask = at.workers.Manager(MyTask(parse={'pos1':1,'pos2':'1'}))
    ... mpTask.run()
    ... print(mpTask._output)
    
Finally, in this quick start, one might want to test the outcome of
running a task and all its subtasks. `apetype.tests` defines the
`TestTask` class that can be inherited from together with, and in
second position, a TaskBase derived class. In the `TestTask` define
all the attributes (without typing) for the test run, together with
the method outcomes as simple attributes.

    >>> class MyTaskTest(at.tests.TestTask, MyTask):
    ...     pos1 = 1
    ...     pos2 = '1'
    ...     addition = 4
    ... at.tests.unittest.main()

"""

from .configs import ConfigBase
from .tasks import TaskBase
